# -*- coding: utf-8 -*-


"""Top-level package for CID (Content IDentifier)."""

__author__ = """Dhruv Baldawa"""
__email__ = "dhruv@dhruvb.com"
__version__ = "0.4.2"

from .cid import CIDv0, CIDv1, make_cid, is_cid, from_string, from_bytes  # noqa: F401
